import os
import shutil
import unittest
import tempfile
from io import open
from fontTools.ufoLib import UFOReader, UFOWriter, UFOLibError
from fontTools.ufoLib import plistlib
from .testSupport import fontInfoVersion2


class TestInfoObject:
    pass


class ReadFontInfoVersion2TestCase(unittest.TestCase):
    def setUp(self):
        self.dstDir = tempfile.mktemp()
        os.mkdir(self.dstDir)
        metaInfo = {"creator": "test", "formatVersion": 2}
        path = os.path.join(self.dstDir, "metainfo.plist")
        with open(path, "wb") as f:
            plistlib.dump(metaInfo, f)

    def tearDown(self):
        shutil.rmtree(self.dstDir)

    def _writeInfoToPlist(self, info):
        path = os.path.join(self.dstDir, "fontinfo.plist")
        with open(path, "wb") as f:
            plistlib.dump(info, f)

    def testRead(self):
        originalData = dict(fontInfoVersion2)
        self._writeInfoToPlist(originalData)
        infoObject = TestInfoObject()
        reader = UFOReader(self.dstDir, validate=True)
        reader.readInfo(infoObject)
        readData = {}
        for attr in list(fontInfoVersion2.keys()):
            readData[attr] = getattr(infoObject, attr)
        self.assertEqual(originalData, readData)

    def testGenericRead(self):
        # familyName
        info = dict(fontInfoVersion2)
        info["familyName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # styleName
        info = dict(fontInfoVersion2)
        info["styleName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # styleMapFamilyName
        info = dict(fontInfoVersion2)
        info["styleMapFamilyName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # styleMapStyleName
        ## not a string
        info = dict(fontInfoVersion2)
        info["styleMapStyleName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## out of range
        info = dict(fontInfoVersion2)
        info["styleMapStyleName"] = "REGULAR"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # versionMajor
        info = dict(fontInfoVersion2)
        info["versionMajor"] = "1"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # versionMinor
        info = dict(fontInfoVersion2)
        info["versionMinor"] = "0"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # copyright
        info = dict(fontInfoVersion2)
        info["copyright"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # trademark
        info = dict(fontInfoVersion2)
        info["trademark"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # unitsPerEm
        info = dict(fontInfoVersion2)
        info["unitsPerEm"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # descender
        info = dict(fontInfoVersion2)
        info["descender"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # xHeight
        info = dict(fontInfoVersion2)
        info["xHeight"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # capHeight
        info = dict(fontInfoVersion2)
        info["capHeight"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # ascender
        info = dict(fontInfoVersion2)
        info["ascender"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # italicAngle
        info = dict(fontInfoVersion2)
        info["italicAngle"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())

    def testHeadRead(self):
        # openTypeHeadCreated
        ## not a string
        info = dict(fontInfoVersion2)
        info["openTypeHeadCreated"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## invalid format
        info = dict(fontInfoVersion2)
        info["openTypeHeadCreated"] = "2000-Jan-01 00:00:00"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeHeadLowestRecPPEM
        info = dict(fontInfoVersion2)
        info["openTypeHeadLowestRecPPEM"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeHeadFlags
        info = dict(fontInfoVersion2)
        info["openTypeHeadFlags"] = [-1]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())

    def testHheaRead(self):
        # openTypeHheaAscender
        info = dict(fontInfoVersion2)
        info["openTypeHheaAscender"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeHheaDescender
        info = dict(fontInfoVersion2)
        info["openTypeHheaDescender"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeHheaLineGap
        info = dict(fontInfoVersion2)
        info["openTypeHheaLineGap"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeHheaCaretSlopeRise
        info = dict(fontInfoVersion2)
        info["openTypeHheaCaretSlopeRise"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeHheaCaretSlopeRun
        info = dict(fontInfoVersion2)
        info["openTypeHheaCaretSlopeRun"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeHheaCaretOffset
        info = dict(fontInfoVersion2)
        info["openTypeHheaCaretOffset"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())

    def testNameRead(self):
        # openTypeNameDesigner
        info = dict(fontInfoVersion2)
        info["openTypeNameDesigner"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameDesignerURL
        info = dict(fontInfoVersion2)
        info["openTypeNameDesignerURL"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameManufacturer
        info = dict(fontInfoVersion2)
        info["openTypeNameManufacturer"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameManufacturerURL
        info = dict(fontInfoVersion2)
        info["openTypeNameManufacturerURL"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameLicense
        info = dict(fontInfoVersion2)
        info["openTypeNameLicense"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameLicenseURL
        info = dict(fontInfoVersion2)
        info["openTypeNameLicenseURL"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameVersion
        info = dict(fontInfoVersion2)
        info["openTypeNameVersion"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameUniqueID
        info = dict(fontInfoVersion2)
        info["openTypeNameUniqueID"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameDescription
        info = dict(fontInfoVersion2)
        info["openTypeNameDescription"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNamePreferredFamilyName
        info = dict(fontInfoVersion2)
        info["openTypeNamePreferredFamilyName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNamePreferredSubfamilyName
        info = dict(fontInfoVersion2)
        info["openTypeNamePreferredSubfamilyName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameCompatibleFullName
        info = dict(fontInfoVersion2)
        info["openTypeNameCompatibleFullName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameSampleText
        info = dict(fontInfoVersion2)
        info["openTypeNameSampleText"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameWWSFamilyName
        info = dict(fontInfoVersion2)
        info["openTypeNameWWSFamilyName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeNameWWSSubfamilyName
        info = dict(fontInfoVersion2)
        info["openTypeNameWWSSubfamilyName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())

    def testOS2Read(self):
        # openTypeOS2WidthClass
        ## not an int
        info = dict(fontInfoVersion2)
        info["openTypeOS2WidthClass"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## out or range
        info = dict(fontInfoVersion2)
        info["openTypeOS2WidthClass"] = 15
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2WeightClass
        info = dict(fontInfoVersion2)
        ## not an int
        info["openTypeOS2WeightClass"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## out of range
        info["openTypeOS2WeightClass"] = -50
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2Selection
        info = dict(fontInfoVersion2)
        info["openTypeOS2Selection"] = [-1]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2VendorID
        info = dict(fontInfoVersion2)
        info["openTypeOS2VendorID"] = 1234
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2Panose
        ## not an int
        info = dict(fontInfoVersion2)
        info["openTypeOS2Panose"] = [0, 1, 2, 3, 4, 5, 6, 7, 8, str(9)]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## too few values
        info = dict(fontInfoVersion2)
        info["openTypeOS2Panose"] = [0, 1, 2, 3]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## too many values
        info = dict(fontInfoVersion2)
        info["openTypeOS2Panose"] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2FamilyClass
        ## not an int
        info = dict(fontInfoVersion2)
        info["openTypeOS2FamilyClass"] = [1, str(1)]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## too few values
        info = dict(fontInfoVersion2)
        info["openTypeOS2FamilyClass"] = [1]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## too many values
        info = dict(fontInfoVersion2)
        info["openTypeOS2FamilyClass"] = [1, 1, 1]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## out of range
        info = dict(fontInfoVersion2)
        info["openTypeOS2FamilyClass"] = [1, 201]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2UnicodeRanges
        ## not an int
        info = dict(fontInfoVersion2)
        info["openTypeOS2UnicodeRanges"] = ["0"]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## out of range
        info = dict(fontInfoVersion2)
        info["openTypeOS2UnicodeRanges"] = [-1]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2CodePageRanges
        ## not an int
        info = dict(fontInfoVersion2)
        info["openTypeOS2CodePageRanges"] = ["0"]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## out of range
        info = dict(fontInfoVersion2)
        info["openTypeOS2CodePageRanges"] = [-1]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2TypoAscender
        info = dict(fontInfoVersion2)
        info["openTypeOS2TypoAscender"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2TypoDescender
        info = dict(fontInfoVersion2)
        info["openTypeOS2TypoDescender"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2TypoLineGap
        info = dict(fontInfoVersion2)
        info["openTypeOS2TypoLineGap"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2WinAscent
        info = dict(fontInfoVersion2)
        info["openTypeOS2WinAscent"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2WinDescent
        info = dict(fontInfoVersion2)
        info["openTypeOS2WinDescent"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2Type
        ## not an int
        info = dict(fontInfoVersion2)
        info["openTypeOS2Type"] = ["1"]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        ## out of range
        info = dict(fontInfoVersion2)
        info["openTypeOS2Type"] = [-1]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2SubscriptXSize
        info = dict(fontInfoVersion2)
        info["openTypeOS2SubscriptXSize"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2SubscriptYSize
        info = dict(fontInfoVersion2)
        info["openTypeOS2SubscriptYSize"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2SubscriptXOffset
        info = dict(fontInfoVersion2)
        info["openTypeOS2SubscriptXOffset"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2SubscriptYOffset
        info = dict(fontInfoVersion2)
        info["openTypeOS2SubscriptYOffset"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2SuperscriptXSize
        info = dict(fontInfoVersion2)
        info["openTypeOS2SuperscriptXSize"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2SuperscriptYSize
        info = dict(fontInfoVersion2)
        info["openTypeOS2SuperscriptYSize"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2SuperscriptXOffset
        info = dict(fontInfoVersion2)
        info["openTypeOS2SuperscriptXOffset"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2SuperscriptYOffset
        info = dict(fontInfoVersion2)
        info["openTypeOS2SuperscriptYOffset"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2StrikeoutSize
        info = dict(fontInfoVersion2)
        info["openTypeOS2StrikeoutSize"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeOS2StrikeoutPosition
        info = dict(fontInfoVersion2)
        info["openTypeOS2StrikeoutPosition"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())

    def testVheaRead(self):
        # openTypeVheaVertTypoAscender
        info = dict(fontInfoVersion2)
        info["openTypeVheaVertTypoAscender"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeVheaVertTypoDescender
        info = dict(fontInfoVersion2)
        info["openTypeVheaVertTypoDescender"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeVheaVertTypoLineGap
        info = dict(fontInfoVersion2)
        info["openTypeVheaVertTypoLineGap"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeVheaCaretSlopeRise
        info = dict(fontInfoVersion2)
        info["openTypeVheaCaretSlopeRise"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeVheaCaretSlopeRun
        info = dict(fontInfoVersion2)
        info["openTypeVheaCaretSlopeRun"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # openTypeVheaCaretOffset
        info = dict(fontInfoVersion2)
        info["openTypeVheaCaretOffset"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())

    def testFONDRead(self):
        # macintoshFONDFamilyID
        info = dict(fontInfoVersion2)
        info["macintoshFONDFamilyID"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # macintoshFONDName
        info = dict(fontInfoVersion2)
        info["macintoshFONDName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())

    def testPostscriptRead(self):
        # postscriptFontName
        info = dict(fontInfoVersion2)
        info["postscriptFontName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # postscriptFullName
        info = dict(fontInfoVersion2)
        info["postscriptFullName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # postscriptSlantAngle
        info = dict(fontInfoVersion2)
        info["postscriptSlantAngle"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, info=TestInfoObject())
        # postscriptUniqueID
        info = dict(fontInfoVersion2)
        info["postscriptUniqueID"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptUnderlineThickness
        info = dict(fontInfoVersion2)
        info["postscriptUnderlineThickness"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptUnderlinePosition
        info = dict(fontInfoVersion2)
        info["postscriptUnderlinePosition"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptIsFixedPitch
        info = dict(fontInfoVersion2)
        info["postscriptIsFixedPitch"] = 2
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptBlueValues
        ## not a list
        info = dict(fontInfoVersion2)
        info["postscriptBlueValues"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## uneven value count
        info = dict(fontInfoVersion2)
        info["postscriptBlueValues"] = [500]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## too many values
        info = dict(fontInfoVersion2)
        info["postscriptBlueValues"] = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptOtherBlues
        ## not a list
        info = dict(fontInfoVersion2)
        info["postscriptOtherBlues"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## uneven value count
        info = dict(fontInfoVersion2)
        info["postscriptOtherBlues"] = [500]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## too many values
        info = dict(fontInfoVersion2)
        info["postscriptOtherBlues"] = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptFamilyBlues
        ## not a list
        info = dict(fontInfoVersion2)
        info["postscriptFamilyBlues"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## uneven value count
        info = dict(fontInfoVersion2)
        info["postscriptFamilyBlues"] = [500]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## too many values
        info = dict(fontInfoVersion2)
        info["postscriptFamilyBlues"] = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptFamilyOtherBlues
        ## not a list
        info = dict(fontInfoVersion2)
        info["postscriptFamilyOtherBlues"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## uneven value count
        info = dict(fontInfoVersion2)
        info["postscriptFamilyOtherBlues"] = [500]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## too many values
        info = dict(fontInfoVersion2)
        info["postscriptFamilyOtherBlues"] = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptStemSnapH
        ## not list
        info = dict(fontInfoVersion2)
        info["postscriptStemSnapH"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## too many values
        info = dict(fontInfoVersion2)
        info["postscriptStemSnapH"] = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptStemSnapV
        ## not list
        info = dict(fontInfoVersion2)
        info["postscriptStemSnapV"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        ## too many values
        info = dict(fontInfoVersion2)
        info["postscriptStemSnapV"] = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptBlueFuzz
        info = dict(fontInfoVersion2)
        info["postscriptBlueFuzz"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptBlueShift
        info = dict(fontInfoVersion2)
        info["postscriptBlueShift"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptBlueScale
        info = dict(fontInfoVersion2)
        info["postscriptBlueScale"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptForceBold
        info = dict(fontInfoVersion2)
        info["postscriptForceBold"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptDefaultWidthX
        info = dict(fontInfoVersion2)
        info["postscriptDefaultWidthX"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptNominalWidthX
        info = dict(fontInfoVersion2)
        info["postscriptNominalWidthX"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptWeightName
        info = dict(fontInfoVersion2)
        info["postscriptWeightName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptDefaultCharacter
        info = dict(fontInfoVersion2)
        info["postscriptDefaultCharacter"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # postscriptWindowsCharacterSet
        info = dict(fontInfoVersion2)
        info["postscriptWindowsCharacterSet"] = -1
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # macintoshFONDFamilyID
        info = dict(fontInfoVersion2)
        info["macintoshFONDFamilyID"] = "abc"
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())
        # macintoshFONDName
        info = dict(fontInfoVersion2)
        info["macintoshFONDName"] = 123
        self._writeInfoToPlist(info)
        reader = UFOReader(self.dstDir, validate=True)
        self.assertRaises(UFOLibError, reader.readInfo, TestInfoObject())


class WriteFontInfoVersion2TestCase(unittest.TestCase):
    def setUp(self):
        self.tempDir = tempfile.mktemp()
        os.mkdir(self.tempDir)
        self.dstDir = os.path.join(self.tempDir, "test.ufo")

    def tearDown(self):
        shutil.rmtree(self.tempDir)

    def makeInfoObject(self):
        infoObject = TestInfoObject()
        for attr, value in list(fontInfoVersion2.items()):
            setattr(infoObject, attr, value)
        return infoObject

    def readPlist(self):
        path = os.path.join(self.dstDir, "fontinfo.plist")
        with open(path, "rb") as f:
            plist = plistlib.load(f)
        return plist

    def testWrite(self):
        infoObject = self.makeInfoObject()
        writer = UFOWriter(self.dstDir, formatVersion=2)
        writer.writeInfo(infoObject)
        writtenData = self.readPlist()
        for attr, originalValue in list(fontInfoVersion2.items()):
            newValue = writtenData[attr]
            self.assertEqual(newValue, originalValue)

    def testGenericWrite(self):
        # familyName
        infoObject = self.makeInfoObject()
        infoObject.familyName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # styleName
        infoObject = self.makeInfoObject()
        infoObject.styleName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # styleMapFamilyName
        infoObject = self.makeInfoObject()
        infoObject.styleMapFamilyName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # styleMapStyleName
        ## not a string
        infoObject = self.makeInfoObject()
        infoObject.styleMapStyleName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## out of range
        infoObject = self.makeInfoObject()
        infoObject.styleMapStyleName = "REGULAR"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # versionMajor
        infoObject = self.makeInfoObject()
        infoObject.versionMajor = "1"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # versionMinor
        infoObject = self.makeInfoObject()
        infoObject.versionMinor = "0"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # copyright
        infoObject = self.makeInfoObject()
        infoObject.copyright = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # trademark
        infoObject = self.makeInfoObject()
        infoObject.trademark = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # unitsPerEm
        infoObject = self.makeInfoObject()
        infoObject.unitsPerEm = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # descender
        infoObject = self.makeInfoObject()
        infoObject.descender = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # xHeight
        infoObject = self.makeInfoObject()
        infoObject.xHeight = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # capHeight
        infoObject = self.makeInfoObject()
        infoObject.capHeight = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # ascender
        infoObject = self.makeInfoObject()
        infoObject.ascender = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # italicAngle
        infoObject = self.makeInfoObject()
        infoObject.italicAngle = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)

    def testHeadWrite(self):
        # openTypeHeadCreated
        ## not a string
        infoObject = self.makeInfoObject()
        infoObject.openTypeHeadCreated = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## invalid format
        infoObject = self.makeInfoObject()
        infoObject.openTypeHeadCreated = "2000-Jan-01 00:00:00"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeHeadLowestRecPPEM
        infoObject = self.makeInfoObject()
        infoObject.openTypeHeadLowestRecPPEM = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeHeadFlags
        infoObject = self.makeInfoObject()
        infoObject.openTypeHeadFlags = [-1]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)

    def testHheaWrite(self):
        # openTypeHheaAscender
        infoObject = self.makeInfoObject()
        infoObject.openTypeHheaAscender = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeHheaDescender
        infoObject = self.makeInfoObject()
        infoObject.openTypeHheaDescender = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeHheaLineGap
        infoObject = self.makeInfoObject()
        infoObject.openTypeHheaLineGap = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeHheaCaretSlopeRise
        infoObject = self.makeInfoObject()
        infoObject.openTypeHheaCaretSlopeRise = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeHheaCaretSlopeRun
        infoObject = self.makeInfoObject()
        infoObject.openTypeHheaCaretSlopeRun = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeHheaCaretOffset
        infoObject = self.makeInfoObject()
        infoObject.openTypeHheaCaretOffset = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)

    def testNameWrite(self):
        # openTypeNameDesigner
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameDesigner = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameDesignerURL
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameDesignerURL = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameManufacturer
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameManufacturer = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameManufacturerURL
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameManufacturerURL = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameLicense
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameLicense = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameLicenseURL
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameLicenseURL = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameVersion
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameVersion = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameUniqueID
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameUniqueID = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameDescription
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameDescription = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNamePreferredFamilyName
        infoObject = self.makeInfoObject()
        infoObject.openTypeNamePreferredFamilyName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNamePreferredSubfamilyName
        infoObject = self.makeInfoObject()
        infoObject.openTypeNamePreferredSubfamilyName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameCompatibleFullName
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameCompatibleFullName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameSampleText
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameSampleText = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameWWSFamilyName
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameWWSFamilyName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeNameWWSSubfamilyName
        infoObject = self.makeInfoObject()
        infoObject.openTypeNameWWSSubfamilyName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)

    def testOS2Write(self):
        # openTypeOS2WidthClass
        ## not an int
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2WidthClass = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## out or range
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2WidthClass = 15
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2WeightClass
        infoObject = self.makeInfoObject()
        ## not an int
        infoObject.openTypeOS2WeightClass = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## out of range
        infoObject.openTypeOS2WeightClass = -50
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2Selection
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2Selection = [-1]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2VendorID
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2VendorID = 1234
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2Panose
        ## not an int
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2Panose = [0, 1, 2, 3, 4, 5, 6, 7, 8, str(9)]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too few values
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2Panose = [0, 1, 2, 3]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too many values
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2Panose = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2FamilyClass
        ## not an int
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2FamilyClass = [0, str(1)]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too few values
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2FamilyClass = [1]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too many values
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2FamilyClass = [1, 1, 1]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## out of range
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2FamilyClass = [1, 20]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2UnicodeRanges
        ## not an int
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2UnicodeRanges = ["0"]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## out of range
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2UnicodeRanges = [-1]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2CodePageRanges
        ## not an int
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2CodePageRanges = ["0"]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## out of range
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2CodePageRanges = [-1]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2TypoAscender
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2TypoAscender = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2TypoDescender
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2TypoDescender = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2TypoLineGap
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2TypoLineGap = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2WinAscent
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2WinAscent = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2WinDescent
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2WinDescent = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2Type
        ## not an int
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2Type = ["1"]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## out of range
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2Type = [-1]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2SubscriptXSize
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2SubscriptXSize = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2SubscriptYSize
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2SubscriptYSize = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2SubscriptXOffset
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2SubscriptXOffset = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2SubscriptYOffset
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2SubscriptYOffset = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2SuperscriptXSize
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2SuperscriptXSize = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2SuperscriptYSize
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2SuperscriptYSize = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2SuperscriptXOffset
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2SuperscriptXOffset = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2SuperscriptYOffset
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2SuperscriptYOffset = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2StrikeoutSize
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2StrikeoutSize = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeOS2StrikeoutPosition
        infoObject = self.makeInfoObject()
        infoObject.openTypeOS2StrikeoutPosition = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)

    def testVheaWrite(self):
        # openTypeVheaVertTypoAscender
        infoObject = self.makeInfoObject()
        infoObject.openTypeVheaVertTypoAscender = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeVheaVertTypoDescender
        infoObject = self.makeInfoObject()
        infoObject.openTypeVheaVertTypoDescender = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeVheaVertTypoLineGap
        infoObject = self.makeInfoObject()
        infoObject.openTypeVheaVertTypoLineGap = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeVheaCaretSlopeRise
        infoObject = self.makeInfoObject()
        infoObject.openTypeVheaCaretSlopeRise = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeVheaCaretSlopeRun
        infoObject = self.makeInfoObject()
        infoObject.openTypeVheaCaretSlopeRun = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # openTypeVheaCaretOffset
        infoObject = self.makeInfoObject()
        infoObject.openTypeVheaCaretOffset = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)

    def testFONDWrite(self):
        # macintoshFONDFamilyID
        infoObject = self.makeInfoObject()
        infoObject.macintoshFONDFamilyID = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # macintoshFONDName
        infoObject = self.makeInfoObject()
        infoObject.macintoshFONDName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)

    def testPostscriptWrite(self):
        # postscriptFontName
        infoObject = self.makeInfoObject()
        infoObject.postscriptFontName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptFullName
        infoObject = self.makeInfoObject()
        infoObject.postscriptFullName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptSlantAngle
        infoObject = self.makeInfoObject()
        infoObject.postscriptSlantAngle = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptUniqueID
        infoObject = self.makeInfoObject()
        infoObject.postscriptUniqueID = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptUnderlineThickness
        infoObject = self.makeInfoObject()
        infoObject.postscriptUnderlineThickness = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptUnderlinePosition
        infoObject = self.makeInfoObject()
        infoObject.postscriptUnderlinePosition = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptIsFixedPitch
        infoObject = self.makeInfoObject()
        infoObject.postscriptIsFixedPitch = 2
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptBlueValues
        ## not a list
        infoObject = self.makeInfoObject()
        infoObject.postscriptBlueValues = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## uneven value count
        infoObject = self.makeInfoObject()
        infoObject.postscriptBlueValues = [500]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too many values
        infoObject = self.makeInfoObject()
        infoObject.postscriptBlueValues = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptOtherBlues
        ## not a list
        infoObject = self.makeInfoObject()
        infoObject.postscriptOtherBlues = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## uneven value count
        infoObject = self.makeInfoObject()
        infoObject.postscriptOtherBlues = [500]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too many values
        infoObject = self.makeInfoObject()
        infoObject.postscriptOtherBlues = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptFamilyBlues
        ## not a list
        infoObject = self.makeInfoObject()
        infoObject.postscriptFamilyBlues = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## uneven value count
        infoObject = self.makeInfoObject()
        infoObject.postscriptFamilyBlues = [500]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too many values
        infoObject = self.makeInfoObject()
        infoObject.postscriptFamilyBlues = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptFamilyOtherBlues
        ## not a list
        infoObject = self.makeInfoObject()
        infoObject.postscriptFamilyOtherBlues = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## uneven value count
        infoObject = self.makeInfoObject()
        infoObject.postscriptFamilyOtherBlues = [500]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too many values
        infoObject = self.makeInfoObject()
        infoObject.postscriptFamilyOtherBlues = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptStemSnapH
        ## not list
        infoObject = self.makeInfoObject()
        infoObject.postscriptStemSnapH = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too many values
        infoObject = self.makeInfoObject()
        infoObject.postscriptStemSnapH = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptStemSnapV
        ## not list
        infoObject = self.makeInfoObject()
        infoObject.postscriptStemSnapV = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        ## too many values
        infoObject = self.makeInfoObject()
        infoObject.postscriptStemSnapV = [
            10,
            20,
            30,
            40,
            50,
            60,
            70,
            80,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
        ]
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptBlueFuzz
        infoObject = self.makeInfoObject()
        infoObject.postscriptBlueFuzz = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptBlueShift
        infoObject = self.makeInfoObject()
        infoObject.postscriptBlueShift = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptBlueScale
        infoObject = self.makeInfoObject()
        infoObject.postscriptBlueScale = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptForceBold
        infoObject = self.makeInfoObject()
        infoObject.postscriptForceBold = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptDefaultWidthX
        infoObject = self.makeInfoObject()
        infoObject.postscriptDefaultWidthX = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptNominalWidthX
        infoObject = self.makeInfoObject()
        infoObject.postscriptNominalWidthX = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptWeightName
        infoObject = self.makeInfoObject()
        infoObject.postscriptWeightName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptDefaultCharacter
        infoObject = self.makeInfoObject()
        infoObject.postscriptDefaultCharacter = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # postscriptWindowsCharacterSet
        infoObject = self.makeInfoObject()
        infoObject.postscriptWindowsCharacterSet = -1
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # macintoshFONDFamilyID
        infoObject = self.makeInfoObject()
        infoObject.macintoshFONDFamilyID = "abc"
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)
        # macintoshFONDName
        infoObject = self.makeInfoObject()
        infoObject.macintoshFONDName = 123
        writer = UFOWriter(self.dstDir, formatVersion=2)
        self.assertRaises(UFOLibError, writer.writeInfo, info=infoObject)



import re
import shutil
from pathlib import Path

import pytest
from fontTools.designspaceLib import (
    AxisDescriptor,
    AxisLabelDescriptor,
    DesignSpaceDocument,
    DiscreteAxisDescriptor,
    InstanceDescriptor,
    LocationLabelDescriptor,
    RangeAxisSubsetDescriptor,
    SourceDescriptor,
    ValueAxisSubsetDescriptor,
    VariableFontDescriptor,
    posix,
)

from .fixtures import datadir


def assert_descriptors_equal(actual, expected):
    assert len(actual) == len(expected)
    for a, e in zip(actual, expected):
        assert a.asdict() == e.asdict()


def test_read_v5_document_simple(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5.designspace")

    assert_descriptors_equal(
        doc.axes,
        [
            AxisDescriptor(
                tag="wght",
                name="Weight",
                minimum=200,
                maximum=1000,
                default=200,
                labelNames={"en": "Wéíght", "fa-IR": "قطر"},
                map=[
                    (200, 0),
                    (300, 100),
                    (400, 368),
                    (600, 600),
                    (700, 824),
                    (900, 1000),
                ],
                axisOrdering=None,
                axisLabels=[
                    AxisLabelDescriptor(
                        name="Extra Light",
                        userMinimum=200,
                        userValue=200,
                        userMaximum=250,
                        labelNames={"de": "Extraleicht", "fr": "Extra léger"},
                    ),
                    AxisLabelDescriptor(
                        name="Light", userMinimum=250, userValue=300, userMaximum=350
                    ),
                    AxisLabelDescriptor(
                        name="Regular",
                        userMinimum=350,
                        userValue=400,
                        userMaximum=450,
                        elidable=True,
                    ),
                    AxisLabelDescriptor(
                        name="Semi Bold",
                        userMinimum=450,
                        userValue=600,
                        userMaximum=650,
                    ),
                    AxisLabelDescriptor(
                        name="Bold", userMinimum=650, userValue=700, userMaximum=850
                    ),
                    AxisLabelDescriptor(
                        name="Black", userMinimum=850, userValue=900, userMaximum=900
                    ),
                    AxisLabelDescriptor(
                        name="Regular",
                        userValue=400,
                        linkedUserValue=700,
                        elidable=True,
                    ),
                    AxisLabelDescriptor(
                        name="Bold", userValue=700, linkedUserValue=400
                    ),
                ],
            ),
            AxisDescriptor(
                tag="wdth",
                name="Width",
                minimum=50,
                maximum=150,
                default=100,
                hidden=True,
                labelNames={"fr": "Chasse"},
                map=[(50, 10), (100, 20), (125, 66), (150, 990)],
                axisOrdering=1,
                axisLabels=[
                    AxisLabelDescriptor(name="Condensed", userValue=50),
                    AxisLabelDescriptor(
                        name="Normal", elidable=True, olderSibling=True, userValue=100
                    ),
                    AxisLabelDescriptor(name="Wide", userValue=125),
                    AxisLabelDescriptor(
                        name="Extra Wide", userValue=150, userMinimum=150
                    ),
                ],
            ),
            DiscreteAxisDescriptor(
                tag="ital",
                name="Italic",
                values=[0, 1],
                default=0,
                axisOrdering=None,
                axisLabels=[
                    AxisLabelDescriptor(
                        name="Roman", userValue=0, elidable=True, linkedUserValue=1
                    ),
                    AxisLabelDescriptor(name="Italic", userValue=1),
                ],
            ),
        ],
    )

    assert_descriptors_equal(
        doc.locationLabels,
        [
            LocationLabelDescriptor(
                name="Some Style",
                labelNames={"fr": "Un Style"},
                userLocation={"Weight": 300, "Width": 50, "Italic": 0},
            ),
            LocationLabelDescriptor(
                name="Other", userLocation={"Weight": 700, "Width": 100, "Italic": 1}
            ),
        ],
    )

    assert_descriptors_equal(
        doc.sources,
        [
            SourceDescriptor(
                filename="masters/masterTest1.ufo",
                path=posix(str((datadir / "masters/masterTest1.ufo").resolve())),
                name="master.ufo1",
                layerName=None,
                location={"Italic": 0.0, "Weight": 0.0, "Width": 20.0},
                copyLib=True,
                copyInfo=True,
                copyGroups=False,
                copyFeatures=True,
                muteKerning=False,
                muteInfo=False,
                mutedGlyphNames=["A", "Z"],
                familyName="MasterFamilyName",
                styleName="MasterStyleNameOne",
                localisedFamilyName={"fr": "Montserrat", "ja": "モンセラート"},
            ),
            SourceDescriptor(
                filename="masters/masterTest2.ufo",
                path=posix(str((datadir / "masters/masterTest2.ufo").resolve())),
                name="master.ufo2",
                layerName=None,
                location={"Italic": 0.0, "Weight": 1000.0, "Width": 20.0},
                copyLib=False,
                copyInfo=False,
                copyGroups=False,
                copyFeatures=False,
                muteKerning=True,
                muteInfo=False,
                mutedGlyphNames=[],
                familyName="MasterFamilyName",
                styleName="MasterStyleNameTwo",
                localisedFamilyName={},
            ),
            SourceDescriptor(
                filename="masters/masterTest2.ufo",
                path=posix(str((datadir / "masters/masterTest2.ufo").resolve())),
                name="master.ufo2",
                layerName="supports",
                location={"Italic": 0.0, "Weight": 1000.0, "Width": 20.0},
                copyLib=False,
                copyInfo=False,
                copyGroups=False,
                copyFeatures=False,
                muteKerning=False,
                muteInfo=False,
                mutedGlyphNames=[],
                familyName="MasterFamilyName",
                styleName="Supports",
                localisedFamilyName={},
            ),
            SourceDescriptor(
                filename="masters/masterTest2.ufo",
                path=posix(str((datadir / "masters/masterTest2.ufo").resolve())),
                name="master.ufo3",
                layerName=None,
                location={"Italic": 1.0, "Weight": 0.0, "Width": 100.0},
                copyLib=False,
                copyGroups=False,
                copyFeatures=False,
                muteKerning=False,
                muteInfo=False,
                mutedGlyphNames=[],
                familyName="MasterFamilyName",
                styleName="FauxItalic",
                localisedFamilyName={},
            ),
        ],
    )

    assert_descriptors_equal(
        doc.variableFonts,
        [
            VariableFontDescriptor(
                name="Test_WghtWdth",
                filename="Test_WghtWdth_different_from_name.ttf",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                    RangeAxisSubsetDescriptor(name="Width"),
                ],
                lib={"com.vtt.source": "sources/vtt/Test_WghtWdth.vtt"},
            ),
            VariableFontDescriptor(
                name="Test_Wght",
                axisSubsets=[RangeAxisSubsetDescriptor(name="Weight")],
                lib={"com.vtt.source": "sources/vtt/Test_Wght.vtt"},
            ),
            VariableFontDescriptor(
                name="TestCd_Wght",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                    ValueAxisSubsetDescriptor(name="Width", userValue=0),
                ],
            ),
            VariableFontDescriptor(
                name="TestWd_Wght",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                    ValueAxisSubsetDescriptor(name="Width", userValue=1000),
                ],
            ),
            VariableFontDescriptor(
                name="TestItalic_Wght",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                    ValueAxisSubsetDescriptor(name="Italic", userValue=1),
                ],
            ),
            VariableFontDescriptor(
                name="TestRB_Wght",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(
                        name="Weight", userMinimum=400, userDefault=400, userMaximum=700
                    ),
                    ValueAxisSubsetDescriptor(name="Italic", userValue=0),
                ],
            ),
        ],
    )

    assert_descriptors_equal(
        doc.instances,
        [
            InstanceDescriptor(
                filename="instances/instanceTest1.ufo",
                path=posix(str((datadir / "instances/instanceTest1.ufo").resolve())),
                name="instance.ufo1",
                designLocation={"Weight": 500.0, "Width": 20.0},
                familyName="InstanceFamilyName",
                styleName="InstanceStyleName",
                postScriptFontName="InstancePostscriptName",
                styleMapFamilyName="InstanceStyleMapFamilyName",
                styleMapStyleName="InstanceStyleMapStyleName",
                localisedFamilyName={"fr": "Montserrat", "ja": "モンセラート"},
                localisedStyleName={"fr": "Demigras", "ja": "半ば"},
                localisedStyleMapFamilyName={
                    "de": "Montserrat Halbfett",
                    "ja": "モンセラート SemiBold",
                },
                localisedStyleMapStyleName={"de": "Standard"},
                glyphs={"arrow": {"mute": True, "unicodes": [291, 292, 293]}},
                lib={
                    "com.coolDesignspaceApp.binaryData": b"<binary gunk>",
                    "com.coolDesignspaceApp.specimenText": "Hamburgerwhatever",
                },
            ),
            InstanceDescriptor(
                filename="instances/instanceTest2.ufo",
                path=posix(str((datadir / "instances/instanceTest2.ufo").resolve())),
                name="instance.ufo2",
                designLocation={"Weight": 500.0, "Width": (400.0, 300.0)},
                familyName="InstanceFamilyName",
                styleName="InstanceStyleName",
                postScriptFontName="InstancePostscriptName",
                styleMapFamilyName="InstanceStyleMapFamilyName",
                styleMapStyleName="InstanceStyleMapStyleName",
                glyphs={
                    "arrow": {
                        "unicodes": [101, 201, 301],
                        "note": "A note about this glyph",
                        "instanceLocation": {"Weight": 120.0, "Width": 100.0},
                        "masters": [
                            {
                                "font": "master.ufo1",
                                "location": {"Weight": 20.0, "Width": 20.0},
                                "glyphName": "BB",
                            },
                            {
                                "font": "master.ufo2",
                                "location": {"Weight": 900.0, "Width": 900.0},
                                "glyphName": "CC",
                            },
                        ],
                    },
                    "arrow2": {},
                },
            ),
            InstanceDescriptor(
                locationLabel="Some Style",
            ),
            InstanceDescriptor(
                designLocation={"Weight": 600.0, "Width": (401.0, 420.0)},
            ),
            InstanceDescriptor(
                designLocation={"Weight": 10.0, "Italic": 0.0},
                userLocation={"Width": 100.0},
            ),
            InstanceDescriptor(
                userLocation={"Weight": 300.0, "Width": 130.0, "Italic": 1.0},
            ),
        ],
    )


def test_read_v5_document_decovar(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5_decovar.designspace")

    assert not doc.variableFonts

    assert_descriptors_equal(
        doc.axes,
        [
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Inline", tag="BLDA"
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Shearded", tag="TRMD"
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Rounded Slab", tag="TRMC"
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Stripes", tag="SKLD"
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Worm Terminal", tag="TRML"
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Inline Skeleton", tag="SKLA"
            ),
            AxisDescriptor(
                default=0,
                maximum=1000,
                minimum=0,
                name="Open Inline Terminal",
                tag="TRMF",
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Inline Terminal", tag="TRMK"
            ),
            AxisDescriptor(default=0, maximum=1000, minimum=0, name="Worm", tag="BLDB"),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Weight", tag="WMX2"
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Flared", tag="TRMB"
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Rounded", tag="TRMA"
            ),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Worm Skeleton", tag="SKLB"
            ),
            AxisDescriptor(default=0, maximum=1000, minimum=0, name="Slab", tag="TRMG"),
            AxisDescriptor(
                default=0, maximum=1000, minimum=0, name="Bifurcated", tag="TRME"
            ),
        ],
    )

    assert_descriptors_equal(
        doc.locationLabels,
        [
            LocationLabelDescriptor(name="Default", elidable=True, userLocation={}),
            LocationLabelDescriptor(
                name="Open", userLocation={"Inline": 1000}, labelNames={"de": "Offen"}
            ),
            LocationLabelDescriptor(name="Worm", userLocation={"Worm": 1000}),
            LocationLabelDescriptor(
                name="Checkered", userLocation={"Inline Skeleton": 1000}
            ),
            LocationLabelDescriptor(
                name="Checkered Reverse", userLocation={"Inline Terminal": 1000}
            ),
            LocationLabelDescriptor(name="Striped", userLocation={"Stripes": 500}),
            LocationLabelDescriptor(name="Rounded", userLocation={"Rounded": 1000}),
            LocationLabelDescriptor(name="Flared", userLocation={"Flared": 1000}),
            LocationLabelDescriptor(
                name="Flared Open",
                userLocation={"Inline Skeleton": 1000, "Flared": 1000},
            ),
            LocationLabelDescriptor(
                name="Rounded Slab", userLocation={"Rounded Slab": 1000}
            ),
            LocationLabelDescriptor(name="Sheared", userLocation={"Shearded": 1000}),
            LocationLabelDescriptor(
                name="Bifurcated", userLocation={"Bifurcated": 1000}
            ),
            LocationLabelDescriptor(
                name="Inline",
                userLocation={"Inline Skeleton": 500, "Open Inline Terminal": 500},
            ),
            LocationLabelDescriptor(name="Slab", userLocation={"Slab": 1000}),
            LocationLabelDescriptor(name="Contrast", userLocation={"Weight": 1000}),
            LocationLabelDescriptor(
                name="Fancy",
                userLocation={"Inline Skeleton": 1000, "Flared": 1000, "Weight": 1000},
            ),
            LocationLabelDescriptor(
                name="Mayhem",
                userLocation={
                    "Inline Skeleton": 1000,
                    "Worm Skeleton": 1000,
                    "Rounded": 500,
                    "Flared": 500,
                    "Rounded Slab": 750,
                    "Bifurcated": 500,
                    "Open Inline Terminal": 250,
                    "Slab": 750,
                    "Inline Terminal": 250,
                    "Worm Terminal": 250,
                    "Weight": 750,
                    "Worm": 1000,
                },
            ),
        ],
    )

    assert [i.locationLabel for i in doc.instances] == [
        "Default",
        "Open",
        "Worm",
        "Checkered",
        "Checkered Reverse",
        "Striped",
        "Rounded",
        "Flared",
        "Flared Open",
        "Rounded Slab",
        "Sheared",
        "Bifurcated",
        "Inline",
        "Slab",
        "Contrast",
        "Fancy",
        "Mayhem",
    ]


def test_read_v5_document_discrete(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5_discrete.designspace")

    assert not doc.locationLabels
    assert not doc.variableFonts

    assert_descriptors_equal(
        doc.axes,
        [
            DiscreteAxisDescriptor(
                default=400,
                values=[400, 700, 900],
                name="Weight",
                tag="wght",
                axisLabels=[
                    AxisLabelDescriptor(
                        name="Regular",
                        userValue=400,
                        elidable=True,
                        linkedUserValue=700,
                    ),
                    AxisLabelDescriptor(name="Bold", userValue=700),
                    AxisLabelDescriptor(name="Black", userValue=900),
                ],
            ),
            DiscreteAxisDescriptor(
                default=100,
                values=[75, 100],
                name="Width",
                tag="wdth",
                axisLabels=[
                    AxisLabelDescriptor(name="Narrow", userValue=75),
                    AxisLabelDescriptor(name="Normal", userValue=100, elidable=True),
                ],
            ),
            DiscreteAxisDescriptor(
                default=0,
                values=[0, 1],
                name="Italic",
                tag="ital",
                axisLabels=[
                    AxisLabelDescriptor(
                        name="Roman", userValue=0, elidable=True, linkedUserValue=1
                    ),
                    AxisLabelDescriptor(name="Italic", userValue=1),
                ],
            ),
        ],
    )


def test_read_v5_document_aktiv(datadir):
    doc = DesignSpaceDocument.fromfile(datadir / "test_v5_aktiv.designspace")

    assert not doc.locationLabels

    assert_descriptors_equal(
        doc.axes,
        [
            AxisDescriptor(
                tag="wght",
                name="Weight",
                minimum=100,
                default=400,
                maximum=900,
                map=[
                    (100, 22),
                    (200, 38),
                    (300, 57),
                    (400, 84),
                    (500, 98),
                    (600, 115),
                    (700, 133),
                    (800, 158),
                    (900, 185),
                ],
                axisOrdering=1,
                axisLabels=[
                    AxisLabelDescriptor(name="Hair", userValue=100),
                    AxisLabelDescriptor(userValue=200, name="Thin"),
                    AxisLabelDescriptor(userValue=300, name="Light"),
                    AxisLabelDescriptor(
                        userValue=400,
                        name="Regular",
                        elidable=True,
                        linkedUserValue=700,
                    ),
                    AxisLabelDescriptor(userValue=500, name="Medium"),
                    AxisLabelDescriptor(userValue=600, name="SemiBold"),
                    AxisLabelDescriptor(userValue=700, name="Bold"),
                    AxisLabelDescriptor(userValue=800, name="XBold"),
                    AxisLabelDescriptor(userValue=900, name="Black"),
                ],
            ),
            AxisDescriptor(
                tag="wdth",
                name="Width",
                minimum=75,
                default=100,
                maximum=125,
                axisOrdering=0,
                axisLabels=[
                    AxisLabelDescriptor(name="Cd", userValue=75),
                    AxisLabelDescriptor(name="Normal", elidable=True, userValue=100),
                    AxisLabelDescriptor(name="Ex", userValue=125),
                ],
            ),
            AxisDescriptor(
                tag="ital",
                name="Italic",
                minimum=0,
                default=0,
                maximum=1,
                axisOrdering=2,
                axisLabels=[
                    AxisLabelDescriptor(
                        name="Upright", userValue=0, elidable=True, linkedUserValue=1
                    ),
                    AxisLabelDescriptor(name="Italic", userValue=1),
                ],
            ),
        ],
    )

    assert_descriptors_equal(
        doc.variableFonts,
        [
            VariableFontDescriptor(
                name="AktivGroteskVF_WghtWdthItal",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                    RangeAxisSubsetDescriptor(name="Width"),
                    RangeAxisSubsetDescriptor(name="Italic"),
                ],
            ),
            VariableFontDescriptor(
                name="AktivGroteskVF_WghtWdth",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                    RangeAxisSubsetDescriptor(name="Width"),
                ],
            ),
            VariableFontDescriptor(
                name="AktivGroteskVF_Wght",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                ],
            ),
            VariableFontDescriptor(
                name="AktivGroteskVF_Italics_WghtWdth",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                    RangeAxisSubsetDescriptor(name="Width"),
                    ValueAxisSubsetDescriptor(name="Italic", userValue=1),
                ],
            ),
            VariableFontDescriptor(
                name="AktivGroteskVF_Italics_Wght",
                axisSubsets=[
                    RangeAxisSubsetDescriptor(name="Weight"),
                    ValueAxisSubsetDescriptor(name="Italic", userValue=1),
                ],
            ),
        ],
    )


@pytest.fixture
def map_doc():
    """Generate a document with a few axes to test the mapping functions"""
    doc = DesignSpaceDocument()
    doc.addAxis(
        AxisDescriptor(
            tag="wght",
            name="Weight",
            minimum=100,
            maximum=900,
            default=100,
            map=[(100, 10), (900, 90)],
        )
    )
    doc.addAxis(
        AxisDescriptor(
            tag="wdth",
            name="Width",
            minimum=75,
            maximum=200,
            default=100,
            map=[(75, 7500), (100, 10000), (200, 20000)],
        )
    )
    doc.addAxis(
        AxisDescriptor(tag="CUST", name="Custom", minimum=1, maximum=2, default=1.5)
    )
    doc.addLocationLabel(
        LocationLabelDescriptor(
            name="Wonky", userLocation={"Weight": 800, "Custom": 1.2}
        )
    )
    return doc


def test_doc_location_map_forward(map_doc: DesignSpaceDocument):
    assert map_doc.map_forward({"Weight": 400, "Width": 150, "Custom": 2}) == {
        "Weight": 40,
        "Width": 15000,
        "Custom": 2,
    }, "The mappings should be used to compute the design locations"
    assert map_doc.map_forward({"Weight": 400}) == {
        "Weight": 40,
        "Width": 10000,
        "Custom": 1.5,
    }, "Missing user locations should be assumed equal to the axis's default"


def test_doc_location_map_backward(map_doc: DesignSpaceDocument):
    assert map_doc.map_backward({"Weight": 40, "Width": 15000, "Custom": 2}) == {
        "Weight": 400,
        "Width": 150,
        "Custom": 2,
    }, "The mappings should be used to compute the user locations"
    assert map_doc.map_backward({"Weight": 40}) == {
        "Weight": 400,
        "Width": 100,
        "Custom": 1.5,
    }, "Missing design locations should be assumed equal to the axis's default"
    assert map_doc.map_backward(
        {"Weight": (40, 50), "Width": (15000, 100000), "Custom": (2, 1.5)}
    ) == {
        "Weight": 400,
        "Width": 150,
        "Custom": 2,
    }, "Only the xvalue of anisotropic locations is used"


def test_instance_location_from_label(map_doc):
    inst = InstanceDescriptor(locationLabel="Wonky")
    assert inst.getFullUserLocation(map_doc) == {
        "Weight": 800,
        "Width": 100,
        "Custom": 1.2,
    }, "an instance with a locationLabel uses the user location from that label, empty values on the label use axis defaults"
    assert inst.getFullDesignLocation(map_doc) == {
        "Weight": 80,
        "Width": 10000,
        "Custom": 1.2,
    }, "an instance with a locationLabel computes the design location from that label, empty values on the label use axis defaults"

    inst = InstanceDescriptor(locationLabel="Wonky", userLocation={"Width": 200})
    assert inst.getFullUserLocation(map_doc) == {
        "Weight": 800,
        "Width": 100,
        "Custom": 1.2,
    }, "an instance with a locationLabel uses the user location from that label, other location values are ignored"
    assert inst.getFullDesignLocation(map_doc) == {
        "Weight": 80,
        "Width": 10000,
        "Custom": 1.2,
    }, "an instance with a locationLabel computes the design location from that label, other location values are ignored"


def test_instance_location_no_data(map_doc):
    inst = InstanceDescriptor()
    assert inst.getFullUserLocation(map_doc) == {
        "Weight": 100,
        "Width": 100,
        "Custom": 1.5,
    }, "an instance without any location data has the default user location"
    assert inst.getFullDesignLocation(map_doc) == {
        "Weight": 10,
        "Width": 10000,
        "Custom": 1.5,
    }, "an instance without any location data has the default design location"


def test_instance_location_design_first(map_doc):
    inst = InstanceDescriptor(
        designLocation={"Weight": (60, 61), "Width": 11000, "Custom": 1.2},
        userLocation={"Weight": 700, "Width": 180, "Custom": 1.4},
    )
    assert inst.getFullUserLocation(map_doc) == {
        "Weight": 600,
        "Width": 110,
        "Custom": 1.2,
    }, "when both design and user location data are provided, design wins"
    assert inst.getFullDesignLocation(map_doc) == {
        "Weight": (60, 61),
        "Width": 11000,
        "Custom": 1.2,
    }, "when both design and user location data are provided, design wins (incl. anisotropy)"


def test_instance_location_mix(map_doc):
    inst = InstanceDescriptor(
        designLocation={"Weight": (60, 61)},
        userLocation={"Width": 180},
    )
    assert inst.getFullUserLocation(map_doc) == {
        "Weight": 600,
        "Width": 180,
        "Custom": 1.5,
    }, "instance location is a mix of design and user locations"
    assert inst.getFullDesignLocation(map_doc) == {
        "Weight": (60, 61),
        "Width": 18000,
        "Custom": 1.5,
    }, "instance location is a mix of design and user location"


@pytest.mark.parametrize(
    "filename",
    [
        "test_v4_original.designspace",
        "test_v5_original.designspace",
        "test_v5_aktiv.designspace",
        "test_v5_decovar.designspace",
        "test_v5_discrete.designspace",
        "test_v5_sourceserif.designspace",
        "test_v5.designspace",
    ],
)
def test_roundtrip(tmpdir, datadir, filename):
    test_file = datadir / filename
    output_path = tmpdir / filename
    # Move the file to the tmpdir so that the filenames stay the same
    # (they're relative to the file's path)
    shutil.copy(test_file, output_path)
    doc = DesignSpaceDocument.fromfile(output_path)
    doc.write(output_path)
    # The input XML has comments and empty lines for documentation purposes
    xml = test_file.read_text(encoding="utf-8")
    xml = re.sub(
        r"<!-- ROUNDTRIP_TEST_REMOVE_ME_BEGIN -->(.|\n)*?<!-- ROUNDTRIP_TEST_REMOVE_ME_END -->",
        "",
        xml,
    )
    xml = re.sub(r"<!--(.|\n)*?-->", "", xml)
    xml = re.sub(r"\s*\n+", "\n", xml)
    assert output_path.read_text(encoding="utf-8") == xml


def test_using_v5_features_upgrades_format(tmpdir, datadir):
    test_file = datadir / "test_v4_original.designspace"
    output_4_path = tmpdir / "test_v4.designspace"
    output_5_path = tmpdir / "test_v5.designspace"
    shutil.copy(test_file, output_4_path)
    doc = DesignSpaceDocument.fromfile(output_4_path)
    doc.write(output_4_path)
    assert 'format="4.1"' in output_4_path.read_text(encoding="utf-8")
    doc.addVariableFont(VariableFontDescriptor(name="TestVF"))
    doc.write(output_5_path)
    assert 'format="5.0"' in output_5_path.read_text(encoding="utf-8")


def test_addAxisDescriptor_discrete():
    ds = DesignSpaceDocument()

    axis = ds.addAxisDescriptor(
        name="Italic",
        tag="ital",
        values=[0, 1],
        default=0,
        hidden=True,
        map=[(0, -12), (1, 0)],
        axisOrdering=3,
        axisLabels=[
            AxisLabelDescriptor(
                name="Roman",
                userValue=0,
                elidable=True,
                olderSibling=True,
                linkedUserValue=1,
                labelNames={"fr": "Romain"},
            )
        ],
    )

    assert ds.axes[0] is axis
    assert_descriptors_equal(
        [axis],
        [
            DiscreteAxisDescriptor(
                tag="ital",
                name="Italic",
                values=[0, 1],
                default=0,
                hidden=True,
                map=[(0, -12), (1, 0)],
                axisOrdering=3,
                axisLabels=[
                    AxisLabelDescriptor(
                        name="Roman",
                        userValue=0,
                        elidable=True,
                        olderSibling=True,
                        linkedUserValue=1,
                        labelNames={"fr": "Romain"},
                    )
                ],
            )
        ],
    )


def test_addLocationLabelDescriptor():
    ds = DesignSpaceDocument()

    label = ds.addLocationLabelDescriptor(
        name="Somewhere",
        userLocation={},
        elidable=True,
        olderSibling=True,
        labelNames={"fr": "Quelque part"},
    )

    assert ds.locationLabels[0] is label
    assert_descriptors_equal(
        [label],
        [
            LocationLabelDescriptor(
                name="Somewhere",
                userLocation={},
                elidable=True,
                olderSibling=True,
                labelNames={"fr": "Quelque part"},
            )
        ],
    )


def test_addVariableFontDescriptor():
    ds = DesignSpaceDocument()

    vf = ds.addVariableFontDescriptor(name="TestVF", filename="TestVF.ttf")

    assert ds.variableFonts[0] is vf
    assert_descriptors_equal(
        [vf], [VariableFontDescriptor(name="TestVF", filename="TestVF.ttf")]
    )



from fontTools.misc.loggingTools import CapturingLogHandler
from fontTools.misc.testTools import parseXML, getXML
from fontTools.misc.textTools import deHexStr
from fontTools.ttLib import TTFont, newTable
from fontTools.misc.fixedTools import log
import os
import unittest


CURR_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DATA_DIR = os.path.join(CURR_DIR, "data")

HHEA_DATA = deHexStr(
    "0001 0000 "  # 1.0   version
    "02EE "  # 750   ascent
    "FF06 "  # -250  descent
    "00C8 "  # 200   lineGap
    "03E8 "  # 1000  advanceWidthMax
    "FFE7 "  # -25   minLeftSideBearing
    "FFEC "  # -20   minRightSideBearing
    "03D1 "  # 977   xMaxExtent
    "0000 "  # 0     caretSlopeRise
    "0001 "  # 1     caretSlopeRun
    "0010 "  # 16    caretOffset
    "0000 "  # 0     reserved0
    "0000 "  # 0     reserved1
    "0000 "  # 0     reserved2
    "0000 "  # 0     reserved3
    "0000 "  # 0     metricDataFormat
    "002A "  # 42    numberOfHMetrics
)

HHEA_AS_DICT = {
    "tableTag": "hhea",
    "tableVersion": 0x00010000,
    "ascent": 750,
    "descent": -250,
    "lineGap": 200,
    "advanceWidthMax": 1000,
    "minLeftSideBearing": -25,
    "minRightSideBearing": -20,
    "xMaxExtent": 977,
    "caretSlopeRise": 0,
    "caretSlopeRun": 1,
    "caretOffset": 16,
    "reserved0": 0,
    "reserved1": 0,
    "reserved2": 0,
    "reserved3": 0,
    "metricDataFormat": 0,
    "numberOfHMetrics": 42,
}

HHEA_XML = [
    '<tableVersion value="0x00010000"/>',
    '<ascent value="750"/>',
    '<descent value="-250"/>',
    '<lineGap value="200"/>',
    '<advanceWidthMax value="1000"/>',
    '<minLeftSideBearing value="-25"/>',
    '<minRightSideBearing value="-20"/>',
    '<xMaxExtent value="977"/>',
    '<caretSlopeRise value="0"/>',
    '<caretSlopeRun value="1"/>',
    '<caretOffset value="16"/>',
    '<reserved0 value="0"/>',
    '<reserved1 value="0"/>',
    '<reserved2 value="0"/>',
    '<reserved3 value="0"/>',
    '<metricDataFormat value="0"/>',
    '<numberOfHMetrics value="42"/>',
]

HHEA_XML_VERSION_AS_FLOAT = [
    '<tableVersion value="1.0"/>',
] + HHEA_XML[1:]


class HheaCompileOrToXMLTest(unittest.TestCase):
    def setUp(self):
        hhea = newTable("hhea")
        hhea.tableVersion = 0x00010000
        hhea.ascent = 750
        hhea.descent = -250
        hhea.lineGap = 200
        hhea.advanceWidthMax = 1000
        hhea.minLeftSideBearing = -25
        hhea.minRightSideBearing = -20
        hhea.xMaxExtent = 977
        hhea.caretSlopeRise = 0
        hhea.caretSlopeRun = 1
        hhea.caretOffset = 16
        hhea.metricDataFormat = 0
        hhea.numberOfHMetrics = 42
        hhea.reserved0 = hhea.reserved1 = hhea.reserved2 = hhea.reserved3 = 0
        self.font = TTFont(sfntVersion="OTTO")
        self.font["hhea"] = hhea

    def test_compile(self):
        hhea = self.font["hhea"]
        hhea.tableVersion = 0x00010000
        self.assertEqual(HHEA_DATA, hhea.compile(self.font))

    def test_compile_version_10_as_float(self):
        hhea = self.font["hhea"]
        hhea.tableVersion = 1.0
        with CapturingLogHandler(log, "WARNING") as captor:
            self.assertEqual(HHEA_DATA, hhea.compile(self.font))
        self.assertTrue(
            len(
                [r for r in captor.records if "Table version value is a float" in r.msg]
            )
            == 1
        )

    def test_toXML(self):
        hhea = self.font["hhea"]
        self.font["hhea"].tableVersion = 0x00010000
        self.assertEqual(getXML(hhea.toXML), HHEA_XML)

    def test_toXML_version_as_float(self):
        hhea = self.font["hhea"]
        hhea.tableVersion = 1.0
        with CapturingLogHandler(log, "WARNING") as captor:
            self.assertEqual(getXML(hhea.toXML), HHEA_XML)
        self.assertTrue(
            len(
                [r for r in captor.records if "Table version value is a float" in r.msg]
            )
            == 1
        )

    def test_aliases(self):
        hhea = self.font["hhea"]
        self.assertEqual(hhea.ascent, hhea.ascender)
        self.assertEqual(hhea.descent, hhea.descender)
        hhea.ascender = 800
        self.assertEqual(hhea.ascent, 800)
        hhea.ascent = 750
        self.assertEqual(hhea.ascender, 750)
        hhea.descender = -300
        self.assertEqual(hhea.descent, -300)
        hhea.descent = -299
        self.assertEqual(hhea.descender, -299)


class HheaDecompileOrFromXMLTest(unittest.TestCase):
    def setUp(self):
        hhea = newTable("hhea")
        self.font = TTFont(sfntVersion="OTTO")
        self.font["hhea"] = hhea

    def test_decompile(self):
        hhea = self.font["hhea"]
        hhea.decompile(HHEA_DATA, self.font)
        for key in hhea.__dict__:
            self.assertEqual(getattr(hhea, key), HHEA_AS_DICT[key])

    def test_fromXML(self):
        hhea = self.font["hhea"]
        for name, attrs, content in parseXML(HHEA_XML):
            hhea.fromXML(name, attrs, content, self.font)
        for key in hhea.__dict__:
            self.assertEqual(getattr(hhea, key), HHEA_AS_DICT[key])

    def test_fromXML_version_as_float(self):
        hhea = self.font["hhea"]
        with CapturingLogHandler(log, "WARNING") as captor:
            for name, attrs, content in parseXML(HHEA_XML_VERSION_AS_FLOAT):
                hhea.fromXML(name, attrs, content, self.font)
        self.assertTrue(
            len(
                [r for r in captor.records if "Table version value is a float" in r.msg]
            )
            == 1
        )
        for key in hhea.__dict__:
            self.assertEqual(getattr(hhea, key), HHEA_AS_DICT[key])


class HheaRecalcTest(unittest.TestCase):
    def test_recalc_TTF(self):
        font = TTFont()
        font.importXML(os.path.join(DATA_DIR, "_h_h_e_a_recalc_TTF.ttx"))
        hhea = font["hhea"]
        hhea.recalc(font)
        self.assertEqual(hhea.advanceWidthMax, 600)
        self.assertEqual(hhea.minLeftSideBearing, -56)
        self.assertEqual(hhea.minRightSideBearing, 100)
        self.assertEqual(hhea.xMaxExtent, 400)

    def test_recalc_OTF(self):
        font = TTFont()
        font.importXML(os.path.join(DATA_DIR, "_h_h_e_a_recalc_OTF.ttx"))
        hhea = font["hhea"]
        hhea.recalc(font)
        self.assertEqual(hhea.advanceWidthMax, 600)
        self.assertEqual(hhea.minLeftSideBearing, -56)
        self.assertEqual(hhea.minRightSideBearing, 100)
        self.assertEqual(hhea.xMaxExtent, 400)

    def test_recalc_empty(self):
        font = TTFont()
        font.importXML(os.path.join(DATA_DIR, "_h_h_e_a_recalc_empty.ttx"))
        hhea = font["hhea"]
        hhea.recalc(font)
        self.assertEqual(hhea.advanceWidthMax, 600)
        self.assertEqual(hhea.minLeftSideBearing, 0)
        self.assertEqual(hhea.minRightSideBearing, 0)
        self.assertEqual(hhea.xMaxExtent, 0)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from fontTools.misc import sstruct
from fontTools.misc.textTools import byteord, safeEval
from . import DefaultTable
import pdb
import struct


METAHeaderFormat = """
		>	# big endian
		tableVersionMajor:			H
		tableVersionMinor:			H
		metaEntriesVersionMajor:	H
		metaEntriesVersionMinor:	H
		unicodeVersion:				L
		metaFlags:					H
		nMetaRecs:					H
"""
# This record is followed by nMetaRecs of METAGlyphRecordFormat.
# This in turn is followd by as many METAStringRecordFormat entries
# as specified by the METAGlyphRecordFormat entries
# this is followed by the strings specifried in the  METAStringRecordFormat
METAGlyphRecordFormat = """
		>	# big endian
		glyphID:			H
		nMetaEntry:			H
"""
# This record is followd by a variable data length field:
# 	USHORT or ULONG	hdrOffset
# Offset from start of META table to the beginning
# of this glyphs array of ns Metadata string entries.
# Size determined by metaFlags field
# METAGlyphRecordFormat entries must be sorted by glyph ID

METAStringRecordFormat = """
		>	# big endian
		labelID:			H
		stringLen:			H
"""
# This record is followd by a variable data length field:
# 	USHORT or ULONG	stringOffset
# METAStringRecordFormat entries must be sorted in order of labelID
# There may be more than one entry with the same labelID
# There may be more than one strign with the same content.

# Strings shall be Unicode UTF-8 encoded, and null-terminated.

METALabelDict = {
    0: "MojikumiX4051",  # An integer in the range 1-20
    1: "UNIUnifiedBaseChars",
    2: "BaseFontName",
    3: "Language",
    4: "CreationDate",
    5: "FoundryName",
    6: "FoundryCopyright",
    7: "OwnerURI",
    8: "WritingScript",
    10: "StrokeCount",
    11: "IndexingRadical",
}


def getLabelString(labelID):
    try:
        label = METALabelDict[labelID]
    except KeyError:
        label = "Unknown label"
    return str(label)


class table_M_E_T_A_(DefaultTable.DefaultTable):
    """Glyphlets META table

    The ``META`` table is used by Adobe's SING Glyphlets.

    See also https://web.archive.org/web/20080627183635/http://www.adobe.com/devnet/opentype/gdk/topic.html
    """

    dependencies = []

    def decompile(self, data, ttFont):
        dummy, newData = sstruct.unpack2(METAHeaderFormat, data, self)
        self.glyphRecords = []
        for i in range(self.nMetaRecs):
            glyphRecord, newData = sstruct.unpack2(
                METAGlyphRecordFormat, newData, GlyphRecord()
            )
            if self.metaFlags == 0:
                [glyphRecord.offset] = struct.unpack(">H", newData[:2])
                newData = newData[2:]
            elif self.metaFlags == 1:
                [glyphRecord.offset] = struct.unpack(">H", newData[:4])
                newData = newData[4:]
            else:
                assert 0, (
                    "The metaFlags field in the META table header has a value other than 0 or 1 :"
                    + str(self.metaFlags)
                )
            glyphRecord.stringRecs = []
            newData = data[glyphRecord.offset :]
            for j in range(glyphRecord.nMetaEntry):
                stringRec, newData = sstruct.unpack2(
                    METAStringRecordFormat, newData, StringRecord()
                )
                if self.metaFlags == 0:
                    [stringRec.offset] = struct.unpack(">H", newData[:2])
                    newData = newData[2:]
                else:
                    [stringRec.offset] = struct.unpack(">H", newData[:4])
                    newData = newData[4:]
                stringRec.string = data[
                    stringRec.offset : stringRec.offset + stringRec.stringLen
                ]
                glyphRecord.stringRecs.append(stringRec)
            self.glyphRecords.append(glyphRecord)

    def compile(self, ttFont):
        offsetOK = 0
        self.nMetaRecs = len(self.glyphRecords)
        count = 0
        while offsetOK != 1:
            count = count + 1
            if count > 4:
                pdb.set_trace()
            metaData = sstruct.pack(METAHeaderFormat, self)
            stringRecsOffset = len(metaData) + self.nMetaRecs * (
                6 + 2 * (self.metaFlags & 1)
            )
            stringRecSize = 6 + 2 * (self.metaFlags & 1)
            for glyphRec in self.glyphRecords:
                glyphRec.offset = stringRecsOffset
                if (glyphRec.offset > 65535) and ((self.metaFlags & 1) == 0):
                    self.metaFlags = self.metaFlags + 1
                    offsetOK = -1
                    break
                metaData = metaData + glyphRec.compile(self)
                stringRecsOffset = stringRecsOffset + (
                    glyphRec.nMetaEntry * stringRecSize
                )
                # this will be the String Record offset for the next GlyphRecord.
            if offsetOK == -1:
                offsetOK = 0
                continue

            # metaData now contains the header and all of the GlyphRecords. Its length should bw
            # the offset to the first StringRecord.
            stringOffset = stringRecsOffset
            for glyphRec in self.glyphRecords:
                assert glyphRec.offset == len(
                    metaData
                ), "Glyph record offset did not compile correctly! for rec:" + str(
                    glyphRec
                )
                for stringRec in glyphRec.stringRecs:
                    stringRec.offset = stringOffset
                    if (stringRec.offset > 65535) and ((self.metaFlags & 1) == 0):
                        self.metaFlags = self.metaFlags + 1
                        offsetOK = -1
                        break
                    metaData = metaData + stringRec.compile(self)
                    stringOffset = stringOffset + stringRec.stringLen
            if offsetOK == -1:
                offsetOK = 0
                continue

            if ((self.metaFlags & 1) == 1) and (stringOffset < 65536):
                self.metaFlags = self.metaFlags - 1
                continue
            else:
                offsetOK = 1

            # metaData now contains the header and all of the GlyphRecords and all of the String Records.
            # Its length should be the offset to the first string datum.
            for glyphRec in self.glyphRecords:
                for stringRec in glyphRec.stringRecs:
                    assert stringRec.offset == len(
                        metaData
                    ), "String offset did not compile correctly! for string:" + str(
                        stringRec.string
                    )
                    metaData = metaData + stringRec.string

        return metaData

    def toXML(self, writer, ttFont):
        writer.comment(
            "Lengths and number of entries in this table will be recalculated by the compiler"
        )
        writer.newline()
        formatstring, names, fixes = sstruct.getformat(METAHeaderFormat)
        for name in names:
            value = getattr(self, name)
            writer.simpletag(name, value=value)
            writer.newline()
        for glyphRec in self.glyphRecords:
            glyphRec.toXML(writer, ttFont)

    def fromXML(self, name, attrs, content, ttFont):
        if name == "GlyphRecord":
            if not hasattr(self, "glyphRecords"):
                self.glyphRecords = []
            glyphRec = GlyphRecord()
            self.glyphRecords.append(glyphRec)
            for element in content:
                if isinstance(element, str):
                    continue
                name, attrs, content = element
                glyphRec.fromXML(name, attrs, content, ttFont)
            glyphRec.offset = -1
            glyphRec.nMetaEntry = len(glyphRec.stringRecs)
        else:
            setattr(self, name, safeEval(attrs["value"]))


class GlyphRecord(object):
    def __init__(self):
        self.glyphID = -1
        self.nMetaEntry = -1
        self.offset = -1
        self.stringRecs = []

    def toXML(self, writer, ttFont):
        writer.begintag("GlyphRecord")
        writer.newline()
        writer.simpletag("glyphID", value=self.glyphID)
        writer.newline()
        writer.simpletag("nMetaEntry", value=self.nMetaEntry)
        writer.newline()
        for stringRec in self.stringRecs:
            stringRec.toXML(writer, ttFont)
        writer.endtag("GlyphRecord")
        writer.newline()

    def fromXML(self, name, attrs, content, ttFont):
        if name == "StringRecord":
            stringRec = StringRecord()
            self.stringRecs.append(stringRec)
            for element in content:
                if isinstance(element, str):
                    continue
                stringRec.fromXML(name, attrs, content, ttFont)
            stringRec.stringLen = len(stringRec.string)
        else:
            setattr(self, name, safeEval(attrs["value"]))

    def compile(self, parentTable):
        data = sstruct.pack(METAGlyphRecordFormat, self)
        if parentTable.metaFlags == 0:
            datum = struct.pack(">H", self.offset)
        elif parentTable.metaFlags == 1:
            datum = struct.pack(">L", self.offset)
        data = data + datum
        return data

    def __repr__(self):
        return (
            "GlyphRecord[ glyphID: "
            + str(self.glyphID)
            + ", nMetaEntry: "
            + str(self.nMetaEntry)
            + ", offset: "
            + str(self.offset)
            + " ]"
        )


# XXX The following two functions are really broken around UTF-8 vs Unicode


def mapXMLToUTF8(string):
    uString = str()
    strLen = len(string)
    i = 0
    while i < strLen:
        prefixLen = 0
        if string[i : i + 3] == "&#x":
            prefixLen = 3
        elif string[i : i + 7] == "&amp;#x":
            prefixLen = 7
        if prefixLen:
            i = i + prefixLen
            j = i
            while string[i] != ";":
                i = i + 1
            valStr = string[j:i]

            uString = uString + chr(eval("0x" + valStr))
        else:
            uString = uString + chr(byteord(string[i]))
        i = i + 1

    return uString.encode("utf_8")


def mapUTF8toXML(string):
    uString = string.decode("utf_8")
    string = ""
    for uChar in uString:
        i = ord(uChar)
        if (i < 0x80) and (i > 0x1F):
            string = string + uChar
        else:
            string = string + "&#x" + hex(i)[2:] + ";"
    return string


class StringRecord(object):
    def toXML(self, writer, ttFont):
        writer.begintag("StringRecord")
        writer.newline()
        writer.simpletag("labelID", value=self.labelID)
        writer.comment(getLabelString(self.labelID))
        writer.newline()
        writer.newline()
        writer.simpletag("string", value=mapUTF8toXML(self.string))
        writer.newline()
        writer.endtag("StringRecord")
        writer.newline()

    def fromXML(self, name, attrs, content, ttFont):
        for element in content:
            if isinstance(element, str):
                continue
            name, attrs, content = element
            value = attrs["value"]
            if name == "string":
                self.string = mapXMLToUTF8(value)
            else:
                setattr(self, name, safeEval(value))

    def compile(self, parentTable):
        data = sstruct.pack(METAStringRecordFormat, self)
        if parentTable.metaFlags == 0:
            datum = struct.pack(">H", self.offset)
        elif parentTable.metaFlags == 1:
            datum = struct.pack(">L", self.offset)
        data = data + datum
        return data

    def __repr__(self):
        return (
            "StringRecord [ labelID: "
            + str(self.labelID)
            + " aka "
            + getLabelString(self.labelID)
            + ", offset: "
            + str(self.offset)
            + ", length: "
            + str(self.stringLen)
            + ", string: "
            + self.string
            + " ]"
        )



from .otBase import BaseTTXConverter


# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6morx.html
class table__m_o_r_x(BaseTTXConverter):
    """The AAT ``morx`` table contains glyph transformations used for script shaping and
    for various other optional smart features, akin to ``GSUB`` and ``GPOS`` features
    in OpenType Layout.

    Note: ``morx`` is a replacement for the now deprecated ``mort`` table.

    See also https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6morx.html
    """

    pass



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Matt Fontaine

from . import E_B_L_C_


class table_C_B_L_C_(E_B_L_C_.table_E_B_L_C_):
    """Color Bitmap Location table

    The ``CBLC`` table contains the locations of color bitmaps for glyphs. It must
    be used in concert with the ``CBDT`` table.

    It is backwards-compatible with the monochrome/grayscale ``EBLC`` table.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/cblc
    """

    dependencies = ["CBDT"]



from fontTools.varLib import _add_avar, load_designspace
from fontTools.varLib.models import VariationModel
from fontTools.varLib.varStore import VarStoreInstancer
from fontTools.misc.fixedTools import fixedToFloat as fi2fl
from fontTools.misc.cliTools import makeOutputFileName
from itertools import product
import logging

log = logging.getLogger("fontTools.varLib.avar")


def _denormalize(v, axis):
    if v >= 0:
        return axis.defaultValue + v * (axis.maxValue - axis.defaultValue)
    else:
        return axis.defaultValue + v * (axis.defaultValue - axis.minValue)


def _pruneLocations(locations, poles, axisTags):
    # Now we have all the input locations, find which ones are
    # not needed and remove them.

    # Note: This algorithm is heavily tied to how VariationModel
    # is implemented.  It assumes that input was extracted from
    # VariationModel-generated object, like an ItemVariationStore
    # created by fontmake using varLib.models.VariationModel.
    # Some CoPilot blabbering:
    # I *think* I can prove that this algorithm is correct, but
    # I'm not 100% sure.  It's possible that there are edge cases
    # where this algorithm will fail.  I'm not sure how to prove
    # that it's correct, but I'm also not sure how to prove that
    # it's incorrect.  I'm not sure how to write a test case that
    # would prove that it's incorrect.  I'm not sure how to write
    # a test case that would prove that it's correct.

    model = VariationModel(locations, axisTags)
    modelMapping = model.mapping
    modelSupports = model.supports
    pins = {tuple(k.items()): None for k in poles}
    for location in poles:
        i = locations.index(location)
        i = modelMapping[i]
        support = modelSupports[i]
        supportAxes = set(support.keys())
        for axisTag, (minV, _, maxV) in support.items():
            for v in (minV, maxV):
                if v in (-1, 0, 1):
                    continue
                for pin in pins.keys():
                    pinLocation = dict(pin)
                    pinAxes = set(pinLocation.keys())
                    if pinAxes != supportAxes:
                        continue
                    if axisTag not in pinAxes:
                        continue
                    if pinLocation[axisTag] == v:
                        break
                else:
                    # No pin found. Go through the previous masters
                    # and find a suitable pin.  Going backwards is
                    # better because it can find a pin that is close
                    # to the pole in more dimensions, and reducing
                    # the total number of pins needed.
                    for candidateIdx in range(i - 1, -1, -1):
                        candidate = modelSupports[candidateIdx]
                        candidateAxes = set(candidate.keys())
                        if candidateAxes != supportAxes:
                            continue
                        if axisTag not in candidateAxes:
                            continue
                        candidate = {
                            k: defaultV for k, (_, defaultV, _) in candidate.items()
                        }
                        if candidate[axisTag] == v:
                            pins[tuple(candidate.items())] = None
                            break
                    else:
                        assert False, "No pin found"
    return [dict(t) for t in pins.keys()]


def mappings_from_avar(font, denormalize=True):
    fvarAxes = font["fvar"].axes
    axisMap = {a.axisTag: a for a in fvarAxes}
    axisTags = [a.axisTag for a in fvarAxes]
    axisIndexes = {a.axisTag: i for i, a in enumerate(fvarAxes)}
    if "avar" not in font:
        return {}, {}
    avar = font["avar"]
    axisMaps = {
        tag: seg
        for tag, seg in avar.segments.items()
        if seg and seg != {-1: -1, 0: 0, 1: 1}
    }
    mappings = []

    if getattr(avar, "majorVersion", 1) == 2:
        varStore = avar.table.VarStore
        regions = varStore.VarRegionList.Region

        # Find all the input locations; this finds "poles", that are
        # locations of the peaks, and "corners", that are locations
        # of the corners of the regions.  These two sets of locations
        # together constitute inputLocations to consider.

        poles = {(): None}  # Just using it as an ordered set
        inputLocations = set({()})
        for varData in varStore.VarData:
            regionIndices = varData.VarRegionIndex
            for regionIndex in regionIndices:
                peakLocation = []
                corners = []
                region = regions[regionIndex]
                for axisIndex, axis in enumerate(region.VarRegionAxis):
                    if axis.PeakCoord == 0:
                        continue
                    axisTag = axisTags[axisIndex]
                    peakLocation.append((axisTag, axis.PeakCoord))
                    corner = []
                    if axis.StartCoord != 0:
                        corner.append((axisTag, axis.StartCoord))
                    if axis.EndCoord != 0:
                        corner.append((axisTag, axis.EndCoord))
                    corners.append(corner)
                corners = set(product(*corners))
                peakLocation = tuple(peakLocation)
                poles[peakLocation] = None
                inputLocations.add(peakLocation)
                inputLocations.update(corners)

        # Sort them by number of axes, then by axis order
        inputLocations = [
            dict(t)
            for t in sorted(
                inputLocations,
                key=lambda t: (len(t), tuple(axisIndexes[tag] for tag, _ in t)),
            )
        ]
        poles = [dict(t) for t in poles.keys()]
        inputLocations = _pruneLocations(inputLocations, list(poles), axisTags)

        # Find the output locations, at input locations
        varIdxMap = avar.table.VarIdxMap
        instancer = VarStoreInstancer(varStore, fvarAxes)
        for location in inputLocations:
            instancer.setLocation(location)
            outputLocation = {}
            for axisIndex, axisTag in enumerate(axisTags):
                varIdx = axisIndex
                if varIdxMap is not None:
                    varIdx = varIdxMap[varIdx]
                delta = instancer[varIdx]
                if delta != 0:
                    v = location.get(axisTag, 0)
                    v = v + fi2fl(delta, 14)
                    # See https://github.com/fonttools/fonttools/pull/3598#issuecomment-2266082009
                    # v = max(-1, min(1, v))
                    outputLocation[axisTag] = v
            mappings.append((location, outputLocation))

        # Remove base master we added, if it maps to the default location
        assert mappings[0][0] == {}
        if mappings[0][1] == {}:
            mappings.pop(0)

    if denormalize:
        for tag, seg in axisMaps.items():
            if tag not in axisMap:
                raise ValueError(f"Unknown axis tag {tag}")
            denorm = lambda v: _denormalize(v, axisMap[tag])
            axisMaps[tag] = {denorm(k): denorm(v) for k, v in seg.items()}

        for i, (inputLoc, outputLoc) in enumerate(mappings):
            inputLoc = {
                tag: _denormalize(val, axisMap[tag]) for tag, val in inputLoc.items()
            }
            outputLoc = {
                tag: _denormalize(val, axisMap[tag]) for tag, val in outputLoc.items()
            }
            mappings[i] = (inputLoc, outputLoc)

    return axisMaps, mappings


def main(args=None):
    """Add `avar` table from designspace file to variable font."""

    if args is None:
        import sys

        args = sys.argv[1:]

    from fontTools import configLogger
    from fontTools.ttLib import TTFont
    from fontTools.designspaceLib import DesignSpaceDocument
    import argparse

    parser = argparse.ArgumentParser(
        "fonttools varLib.avar",
        description="Add `avar` table from designspace file to variable font.",
    )
    parser.add_argument("font", metavar="varfont.ttf", help="Variable-font file.")
    parser.add_argument(
        "designspace",
        metavar="family.designspace",
        help="Designspace file.",
        nargs="?",
        default=None,
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=str,
        help="Output font file name.",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Run more verbosely."
    )

    options = parser.parse_args(args)

    configLogger(level=("INFO" if options.verbose else "WARNING"))

    font = TTFont(options.font)
    if not "fvar" in font:
        log.error("Not a variable font.")
        return 1

    if options.designspace is None:
        from pprint import pprint

        segments, mappings = mappings_from_avar(font)
        pprint(segments)
        pprint(mappings)
        print(len(mappings), "mappings")
        return

    axisTags = [a.axisTag for a in font["fvar"].axes]

    ds = load_designspace(options.designspace, require_sources=False)

    if "avar" in font:
        log.warning("avar table already present, overwriting.")
        del font["avar"]

    _add_avar(font, ds.axes, ds.axisMappings, axisTags)

    if options.output_file is None:
        outfile = makeOutputFileName(options.font, overWrite=True, suffix=".avar")
    else:
        outfile = options.output_file
    if outfile:
        log.info("Saving %s", outfile)
        font.save(outfile)


if __name__ == "__main__":
    import sys

    sys.exit(main())



from .otBase import BaseTTXConverter


class table_B_A_S_E_(BaseTTXConverter):
    """Baseline table

    The ``BASE`` table contains information needed to align glyphs in
    different scripts, from different fonts, or at different sizes
    within the same line of text.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/base
    """

    pass



from fontTools.ttLib import newTable
from fontTools.ttLib.tables._k_e_r_n import KernTable_format_0, KernTable_format_unkown
from fontTools.misc.textTools import deHexStr
from fontTools.misc.testTools import FakeFont, getXML, parseXML
import itertools
import pytest


KERN_VER_0_FMT_0_DATA = deHexStr(
    "0000 "  #  0: version=0
    "0001 "  #  2: nTables=1
    "0000 "  #  4: version=0 (bogus field, unused)
    "0020 "  #  6: length=32
    "00 "  #  8: format=0
    "01 "  #  9: coverage=1
    "0003 "  # 10: nPairs=3
    "000C "  # 12: searchRange=12
    "0001 "  # 14: entrySelector=1
    "0006 "  # 16: rangeShift=6
    "0004 000C FFD8 "  # 18: l=4, r=12, v=-40
    "0004 001C 0028 "  # 24: l=4, r=28, v=40
    "0005 0028 FFCE "  # 30: l=5, r=40, v=-50
)
assert len(KERN_VER_0_FMT_0_DATA) == 36

KERN_VER_0_FMT_0_XML = [
    '<version value="0"/>',
    '<kernsubtable coverage="1" format="0">',
    '  <pair l="E" r="M" v="-40"/>',
    '  <pair l="E" r="c" v="40"/>',
    '  <pair l="F" r="o" v="-50"/>',
    "</kernsubtable>",
]

KERN_VER_1_FMT_0_DATA = deHexStr(
    "0001 0000 "  #  0: version=1
    "0000 0001 "  #  4: nTables=1
    "0000 0022 "  #  8: length=34
    "00 "  # 12: coverage=0
    "00 "  # 13: format=0
    "0000 "  # 14: tupleIndex=0
    "0003 "  # 16: nPairs=3
    "000C "  # 18: searchRange=12
    "0001 "  # 20: entrySelector=1
    "0006 "  # 22: rangeShift=6
    "0004 000C FFD8 "  # 24: l=4, r=12, v=-40
    "0004 001C 0028 "  # 30: l=4, r=28, v=40
    "0005 0028 FFCE "  # 36: l=5, r=40, v=-50
)
assert len(KERN_VER_1_FMT_0_DATA) == 42

KERN_VER_1_FMT_0_XML = [
    '<version value="1.0"/>',
    '<kernsubtable coverage="0" format="0" tupleIndex="0">',
    '  <pair l="E" r="M" v="-40"/>',
    '  <pair l="E" r="c" v="40"/>',
    '  <pair l="F" r="o" v="-50"/>',
    "</kernsubtable>",
]

KERN_VER_0_FMT_UNKNOWN_DATA = deHexStr(
    "0000 "  #  0: version=0
    "0002 "  #  2: nTables=2
    "0000 "  #  4: version=0
    "000A "  #  6: length=10
    "04 "  #  8: format=4  (format 4 doesn't exist)
    "01 "  #  9: coverage=1
    "1234 5678 "  # 10: garbage...
    "0000 "  # 14: version=0
    "000A "  # 16: length=10
    "05 "  # 18: format=5  (format 5 doesn't exist)
    "01 "  # 19: coverage=1
    "9ABC DEF0 "  # 20: garbage...
)
assert len(KERN_VER_0_FMT_UNKNOWN_DATA) == 24

KERN_VER_0_FMT_UNKNOWN_XML = [
    '<version value="0"/>',
    '<kernsubtable format="4">',
    "  <!-- unknown 'kern' subtable format -->",
    "  0000000A 04011234",
    "  5678             ",
    "</kernsubtable>",
    '<kernsubtable format="5">',
    "<!-- unknown 'kern' subtable format -->",
    "  0000000A 05019ABC",
    "  DEF0             ",
    "</kernsubtable>",
]

KERN_VER_1_FMT_UNKNOWN_DATA = deHexStr(
    "0001 0000 "  #  0: version=1
    "0000 0002 "  #  4: nTables=2
    "0000 000C "  #  8: length=12
    "00 "  # 12: coverage=0
    "04 "  # 13: format=4  (format 4 doesn't exist)
    "0000 "  # 14: tupleIndex=0
    "1234 5678"  # 16: garbage...
    "0000 000C "  # 20: length=12
    "00 "  # 24: coverage=0
    "05 "  # 25: format=5  (format 5 doesn't exist)
    "0000 "  # 26: tupleIndex=0
    "9ABC DEF0 "  # 28: garbage...
)
assert len(KERN_VER_1_FMT_UNKNOWN_DATA) == 32

KERN_VER_1_FMT_UNKNOWN_XML = [
    '<version value="1"/>',
    '<kernsubtable format="4">',
    "  <!-- unknown 'kern' subtable format -->",
    "  0000000C 00040000",
    "  12345678         ",
    "</kernsubtable>",
    '<kernsubtable format="5">',
    "  <!-- unknown 'kern' subtable format -->",
    "  0000000C 00050000",
    "  9ABCDEF0         ",
    "</kernsubtable>",
]

KERN_VER_0_FMT_0_OVERFLOWING_DATA = deHexStr(
    "0000 "  #  0: version=0
    "0001 "  #  2: nTables=1
    "0000 "  #  4: version=0 (bogus field, unused)
    "0274 "  #  6: length=628 (bogus value for 66164 % 0x10000)
    "00 "  #  8: format=0
    "01 "  #  9: coverage=1
    "2B11 "  # 10: nPairs=11025
    "C000 "  # 12: searchRange=49152
    "000D "  # 14: entrySelector=13
    "4266 "  # 16: rangeShift=16998
) + deHexStr(
    " ".join(
        "%04X %04X %04X" % (a, b, 0)
        for (a, b) in itertools.product(range(105), repeat=2)
    )
)


@pytest.fixture
def font():
    return FakeFont(list("ABCDEFGHIJKLMNOPQRSTUVWXYZ" "abcdefghijklmnopqrstuvwxyz"))


@pytest.fixture
def overflowing_font():
    return FakeFont(["glyph%i" % i for i in range(105)])


class KernTableTest(object):
    @pytest.mark.parametrize(
        "data, version",
        [
            (KERN_VER_0_FMT_0_DATA, 0),
            (KERN_VER_1_FMT_0_DATA, 1.0),
        ],
        ids=["version_0", "version_1"],
    )
    def test_decompile_single_format_0(self, data, font, version):
        kern = newTable("kern")
        kern.decompile(data, font)

        assert kern.version == version
        assert len(kern.kernTables) == 1

        st = kern.kernTables[0]
        assert st.apple is (version == 1.0)
        assert st.format == 0
        # horizontal kerning in OT kern is coverage 0x01, while in
        # AAT kern it's the default (0)
        assert st.coverage == (0 if st.apple else 1)
        assert st.tupleIndex == (0 if st.apple else None)
        assert len(st.kernTable) == 3
        assert st.kernTable == {("E", "M"): -40, ("E", "c"): 40, ("F", "o"): -50}

    @pytest.mark.parametrize(
        "version, expected",
        [
            (0, KERN_VER_0_FMT_0_DATA),
            (1.0, KERN_VER_1_FMT_0_DATA),
        ],
        ids=["version_0", "version_1"],
    )
    def test_compile_single_format_0(self, font, version, expected):
        kern = newTable("kern")
        kern.version = version
        apple = version == 1.0
        st = KernTable_format_0(apple)
        kern.kernTables = [st]
        st.coverage = 0 if apple else 1
        st.tupleIndex = 0 if apple else None
        st.kernTable = {("E", "M"): -40, ("E", "c"): 40, ("F", "o"): -50}
        data = kern.compile(font)
        assert data == expected

    @pytest.mark.parametrize(
        "xml, version",
        [
            (KERN_VER_0_FMT_0_XML, 0),
            (KERN_VER_1_FMT_0_XML, 1.0),
        ],
        ids=["version_0", "version_1"],
    )
    def test_fromXML_single_format_0(self, xml, font, version):
        kern = newTable("kern")
        for name, attrs, content in parseXML(xml):
            kern.fromXML(name, attrs, content, ttFont=font)

        assert kern.version == version
        assert len(kern.kernTables) == 1

        st = kern.kernTables[0]
        assert st.apple is (version == 1.0)
        assert st.format == 0
        assert st.coverage == (0 if st.apple else 1)
        assert st.tupleIndex == (0 if st.apple else None)
        assert len(st.kernTable) == 3
        assert st.kernTable == {("E", "M"): -40, ("E", "c"): 40, ("F", "o"): -50}

    @pytest.mark.parametrize(
        "version, expected",
        [
            (0, KERN_VER_0_FMT_0_XML),
            (1.0, KERN_VER_1_FMT_0_XML),
        ],
        ids=["version_0", "version_1"],
    )
    def test_toXML_single_format_0(self, font, version, expected):
        kern = newTable("kern")
        kern.version = version
        apple = version == 1.0
        st = KernTable_format_0(apple)
        kern.kernTables = [st]
        st.coverage = 0 if apple else 1
        st.tupleIndex = 0 if apple else None
        st.kernTable = {("E", "M"): -40, ("E", "c"): 40, ("F", "o"): -50}
        xml = getXML(kern.toXML, font)
        assert xml == expected

    @pytest.mark.parametrize(
        "data, version, header_length, st_length",
        [
            (KERN_VER_0_FMT_UNKNOWN_DATA, 0, 4, 10),
            (KERN_VER_1_FMT_UNKNOWN_DATA, 1.0, 8, 12),
        ],
        ids=["version_0", "version_1"],
    )
    def test_decompile_format_unknown(
        self, data, font, version, header_length, st_length
    ):
        kern = newTable("kern")
        kern.decompile(data, font)

        assert kern.version == version
        assert len(kern.kernTables) == 2

        st_data = data[header_length:]
        st0 = kern.kernTables[0]
        assert st0.format == 4
        assert st0.data == st_data[:st_length]
        st_data = st_data[st_length:]

        st1 = kern.kernTables[1]
        assert st1.format == 5
        assert st1.data == st_data[:st_length]

    @pytest.mark.parametrize(
        "version, st_length, expected",
        [
            (0, 10, KERN_VER_0_FMT_UNKNOWN_DATA),
            (1.0, 12, KERN_VER_1_FMT_UNKNOWN_DATA),
        ],
        ids=["version_0", "version_1"],
    )
    def test_compile_format_unknown(self, version, st_length, expected):
        kern = newTable("kern")
        kern.version = version
        kern.kernTables = []

        for unknown_fmt, kern_data in zip((4, 5), ("1234 5678", "9ABC DEF0")):
            if version > 0:
                coverage = 0
                header_fmt = deHexStr(
                    "%08X %02X %02X %04X" % (st_length, coverage, unknown_fmt, 0)
                )
            else:
                coverage = 1
                header_fmt = deHexStr(
                    "%04X %04X %02X %02X" % (0, st_length, unknown_fmt, coverage)
                )
            st = KernTable_format_unkown(unknown_fmt)
            st.data = header_fmt + deHexStr(kern_data)
            kern.kernTables.append(st)

        data = kern.compile(font)
        assert data == expected

    @pytest.mark.parametrize(
        "xml, version, st_length",
        [
            (KERN_VER_0_FMT_UNKNOWN_XML, 0, 10),
            (KERN_VER_1_FMT_UNKNOWN_XML, 1.0, 12),
        ],
        ids=["version_0", "version_1"],
    )
    def test_fromXML_format_unknown(self, xml, font, version, st_length):
        kern = newTable("kern")
        for name, attrs, content in parseXML(xml):
            kern.fromXML(name, attrs, content, ttFont=font)

        assert kern.version == version
        assert len(kern.kernTables) == 2

        st0 = kern.kernTables[0]
        assert st0.format == 4
        assert len(st0.data) == st_length

        st1 = kern.kernTables[1]
        assert st1.format == 5
        assert len(st1.data) == st_length

    @pytest.mark.parametrize("version", [0, 1.0], ids=["version_0", "version_1"])
    def test_toXML_format_unknown(self, font, version):
        kern = newTable("kern")
        kern.version = version
        st = KernTable_format_unkown(4)
        st.data = b"ABCD"
        kern.kernTables = [st]

        xml = getXML(kern.toXML, font)

        assert xml == [
            '<version value="%s"/>' % version,
            '<kernsubtable format="4">',
            "  <!-- unknown 'kern' subtable format -->",
            "  41424344   ",
            "</kernsubtable>",
        ]

    def test_getkern(self):
        table = newTable("kern")
        table.version = 0
        table.kernTables = []

        assert table.getkern(0) is None

        st0 = KernTable_format_0()
        table.kernTables.append(st0)

        assert table.getkern(0) is st0
        assert table.getkern(4) is None

        st1 = KernTable_format_unkown(4)
        table.kernTables.append(st1)


class KernTable_format_0_Test(object):
    def test_decompileBadGlyphId(self, font):
        subtable = KernTable_format_0()
        subtable.decompile(
            b"\x00"
            + b"\x00"
            + b"\x00"
            + b"\x1a"
            + b"\x00"
            + b"\x00"
            + b"\x00"
            + b"\x02"
            + b"\x00" * 6
            + b"\x00"
            + b"\x01"
            + b"\x00"
            + b"\x03"
            + b"\x00"
            + b"\x01"
            + b"\x00"
            + b"\x01"
            + b"\xFF"
            + b"\xFF"
            + b"\x00"
            + b"\x02",
            font,
        )
        assert subtable[("B", "D")] == 1
        assert subtable[("B", "glyph65535")] == 2

    def test_compileOverflowingSubtable(self, overflowing_font):
        font = overflowing_font
        kern = newTable("kern")
        kern.version = 0
        st = KernTable_format_0(0)
        kern.kernTables = [st]
        st.coverage = 1
        st.tupleIndex = None
        st.kernTable = {
            (a, b): 0 for (a, b) in itertools.product(font.getGlyphOrder(), repeat=2)
        }
        assert len(st.kernTable) == 11025
        data = kern.compile(font)
        assert data == KERN_VER_0_FMT_0_OVERFLOWING_DATA

    def test_decompileOverflowingSubtable(self, overflowing_font):
        font = overflowing_font
        data = KERN_VER_0_FMT_0_OVERFLOWING_DATA
        kern = newTable("kern")
        kern.decompile(data, font)

        st = kern.kernTables[0]
        assert st.kernTable == {
            (a, b): 0 for (a, b) in itertools.product(font.getGlyphOrder(), repeat=2)
        }


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(sys.argv))



from fontTools.misc.testTools import parseXmlInto, getXML
from fontTools.ttLib.tables.D__e_b_g import table_D__e_b_g


DEBG_DATA = {
    "com.github.fonttools.feaLib": {
        "GPOS": {"0": ["<features>:6:5", "kern_Default", ["DFLT", "dflt", "kern"]]}
    }
}


DEBG_XML = """\
<json>
  <![CDATA[{
    "com.github.fonttools.feaLib": {
      "GPOS": {
        "0": [
          "<features>:6:5",
          "kern_Default",
          [
            "DFLT",
            "dflt",
            "kern"
          ]
        ]
      }
    }
  }]]>
</json>
"""


def test_compile_decompile_and_roundtrip_ttx():
    ttFont = None
    table = table_D__e_b_g()
    assert table.tableTag == "Debg"
    assert table.data == {}
    table.data = DEBG_DATA

    data = table.compile(ttFont)

    table2 = table_D__e_b_g()
    table2.decompile(data, ttFont)

    assert table.data == table2.data

    assert getXML(table2.toXML) == DEBG_XML.splitlines()

    table3 = table_D__e_b_g()
    parseXmlInto(ttFont, table3, DEBG_XML)

    assert table3.data == DEBG_DATA
